/* -*- Mode: C; c-file-style: "gnu" -*-
   jvmdithr.c -- JVMDI Thread Execution Functions
   Created: Hernan Otero <hernan_otero@bigfoot.com>, 18-Jul-1998.
 */
/*
  This file is part of Japhar, the GNU Virtual Machine for Java Bytecodes.
  Japhar is a project of The Hungry Programmers, GNU, and OryxSoft.

  Copyright (C) 1998 The Hungry Programmers

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Library General Public
  License as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Library General Public License for more details.

  You should have received a copy of the GNU Library General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  */
#include "config.h"

#include "jvmdiint.h"
#include "interp.h"
#include "objects.h"
#include "log.h"

#define MYLOG "JVMDI"

JNIEXPORT jvmdiError JNICALL
JVMDI_GetThreadStatus(JNIEnv *env,
		      jthread thread,
		      jint *statusPtr)
{
  japhar_obj thread_obj = (japhar_obj)thread;
  JThreadInfo *thread_info;

  if (env == NULL
      || thread_obj == NULL
      || statusPtr == NULL)
    return JVMDI_ERROR_NULL_POINTER;

  thread_info = thread_obj_to_jthread(env, thread_obj);
  if (thread_info == NULL)
    return JVMDI_ERROR_INVALID_THREAD;
  
  switch (STATE(thread_info)) {
    case STATE_INTERRUPTED:
	  /*
	   *
	   */
	  *statusPtr = JVMDI_THREAD_STATUS_RUNNING;
	  break;
    case STATE_RUNNING:
	  /*
	   * Thread is runnable.
	   */
	  *statusPtr = JVMDI_THREAD_STATUS_RUNNING;
	  break;
    case STATE_FINISHED:
	  /*
	   * Thread is waiting to die.
	   */
      *statusPtr = JVMDI_THREAD_STATUS_ZOMBIE;
	  break;
#if WITH_JVMDI
	case STATE_SUSPENDED:
      /*
       * Thread suspended. (Thread.suspend(), JVM_Suspend(), or
       * JVMDI_Suspend().) 
       */
      *statusPtr = JVMDI_THREAD_STATUS_SUSPENDED;
	  break;
#endif
#if 0
      /* XXX need to consider these cases too! */
      /*
       * Thread sleeping. (Thread.sleep() or JVM_Sleep().) 
       */
      *statusPtr = JVMDI_THREAD_STATUS_SLEEPING
	  break;

      /*
       * Synchronization block. 
       */
      *statusPtr = JVMDI_THREAD_STATUS_MONITOR 
	  break;

      /*
       * Thread waiting. (Thread.wait() or JVM_MonitorWait().) 
       */
      *statusPtr = JVMDI_THREAD_STATUS_WAIT 
	  break;

      /*
       * Thread at breakpoint.
       */
      *statusPtr = JVMDI_THREAD_STATUS_BREAK;
	  break;
#endif
	default:
	  /*
	   * Status unknown
	   */
      *statusPtr = JVMDI_THREAD_STATUS_UNKNOWN;
	  break;
  }
  return JVMDI_ERROR_NONE;
}

JNIEXPORT jvmdiError JNICALL
JVMDI_GetAllThreads(JNIEnv *env, jint *threadsCountPtr,
                    jthread **threadsPtr)
{
  return JVMDI_ERROR_NOT_IMPLEMENTED;
}

JNIEXPORT jvmdiError JNICALL
JVMDI_SuspendThread(JNIEnv *env,
		    jthread thread)
{
  japhar_obj thread_obj = (japhar_obj)thread;
  JThreadInfo *thread_info;

  if (env == NULL
      || thread_obj == NULL)
    return JVMDI_ERROR_NULL_POINTER;

  thread_info = thread_obj_to_jthread(env, thread_obj);
  if (thread_info == NULL)
    return JVMDI_ERROR_INVALID_THREAD;

#if WITH_JVMDI
  /* the thread cant be suspended for this operation. */
  if (STATE(thread_info) == STATE_SUSPENDED) {
    return JVMDI_ERROR_THREAD_SUSPENDED;
  }
#endif

  /* XXX should check for JVMDI_ERROR_DEAD_VM */

  JAVARLOG1(MYLOG, 1, "suspending thread(%p)\n", thread_obj);
  suspend_thread(thread_obj, thread_info);

  return JVMDI_ERROR_NONE;
}

JNIEXPORT jvmdiError JNICALL
JVMDI_ResumeThread(JNIEnv *env,
		   jthread thread)
{
  japhar_obj thread_obj = (japhar_obj)thread;
  JThreadInfo *thread_info;

  if (env == NULL
      || thread_obj == NULL)
    return JVMDI_ERROR_NULL_POINTER;

  thread_info = thread_obj_to_jthread(env, thread_obj);
  if (thread_info == NULL)
    return JVMDI_ERROR_INVALID_THREAD;

#if WITH_JVMDI
  /* the thread must be suspended for this operation. */
  if (STATE(thread_info) != STATE_SUSPENDED)
    return JVMDI_ERROR_THREAD_NOT_SUSPENDED;
#endif

  /* XXX should check for JVMDI_ERROR_DEAD_VM */

  JAVARLOG1(MYLOG, 1, "resuming thread(%p)\n", thread_obj);
  resume_thread(thread_obj, thread_info);

  return JVMDI_ERROR_NONE;
}

JNIEXPORT jvmdiError JNICALL
JVMDI_StopThread(JNIEnv *env, jthread thread, jobject exception)
{
  return JVMDI_ERROR_NOT_IMPLEMENTED;
}

JNIEXPORT jvmdiError JNICALL
JVMDI_InterruptThread(JNIEnv *env, jthread thread)
{
  return JVMDI_ERROR_NOT_IMPLEMENTED;
}

JNIEXPORT jvmdiError JNICALL
JVMDI_SetSingleStep(JNIEnv *env,
		    jthread thread,
		    jboolean shouldStep)
{
  japhar_obj thread_obj = (japhar_obj)thread;
  JThreadInfo *thread_info;

  if (env == NULL
      || thread_obj == NULL)
    return JVMDI_ERROR_NULL_POINTER;

  thread_info = thread_obj_to_jthread(env, thread_obj);
  if (thread_info == NULL)
    return JVMDI_ERROR_INVALID_THREAD;

#if WITH_JVMDI
  /* the thread must be suspended for this operation. */
  if (STATE(thread_info) != STATE_SUSPENDED)
    return JVMDI_ERROR_THREAD_NOT_SUSPENDED;
#endif

  /* XXX should check for JVMDI_ERROR_DEAD_VM */

  /* XXX should actually do something here! */
  fprintf(stderr, "set single step on thread (%p) ignored - not yet implemented\n", thread_obj);

  return JVMDI_ERROR_NONE;
}
