/* -*- Mode: C++; c-file-style: "gnu" -*-
   japharJVMPlugin.h -- japhar's mozilla java runtime support
   Created: Chris Toshok <toshok@hungry.com>, 8-Aug-1998
 */
/*
  This file is part of Japhar, the GNU Virtual Machine for Java Bytecodes.
  Japhar is a project of The Hungry Programmers, GNU, and OryxSoft.

  Copyright (C) 1997, 1998 The Hungry Programmers

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Library General Public
  License as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Library General Public License for more details.

  You should have received a copy of the GNU Library General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  */

#ifndef _japharJVMPlugin_h
#define _japharJVMPlugin_h

/* nscp headers */
#include "nsplugin.h"
#include "nsIJVMPlugin.h"

/* japhar headers */
#include "jni.h"
#include "native-threads.h"
#include "japharJVMConsole.h"

class japharJVMPlugin : public nsIJVMPlugin {
public:

  japharJVMPlugin();

  /////////////////////////////////////////////////
  // nsISupports
  /////////////////////////////////////////////////
  NS_DECL_ISUPPORTS

  /////////////////////////////////////////////////
  // nsIFactory
  /////////////////////////////////////////////////

  NS_IMETHOD CreateInstance(nsISupports *aOuter,
                            REFNSIID aIID,
                            void **aResult);

  NS_IMETHOD LockFactory(PRBool aLock);

  /////////////////////////////////////////////////
  // nsIPlugin
  /////////////////////////////////////////////////

  /**
   * Initializes the plugin and will be called before any new instances are
   * created. It is passed browserInterfaces on which QueryInterface
   * may be used to obtain an nsIPluginManager, and other interfaces.
   *
   * @param browserInterfaces - an object that allows access to other browser
   * interfaces via QueryInterface
   * @result - NS_OK if this operation was successful
   */
  NS_IMETHOD
  Initialize(nsISupports* browserInterfaces);
  
  /**
   * Called when the browser is done with the plugin factory, or when
   * the plugin is disabled by the user.
   *
   * (Corresponds to NPP_Shutdown.)
   *
   * @result - NS_OK if this operation was successful
   */
  NS_IMETHOD
  Shutdown(void);

  /**
   * Returns the MIME description for the plugin. The MIME description 
   * is a colon-separated string containg the plugin MIME type, plugin
   * data file extension, and plugin name, e.g.:
   *
   * "application/x-simple-plugin:smp:Simple LiveConnect Sample Plug-in"
   *
   * (Corresponds to NPP_GetMIMEDescription.)
   *
   * @param resultingDesc - the resulting MIME description 
   * @result - NS_OK if this operation was successful
   */
  NS_IMETHOD
  GetMIMEDescription(const char* *resultingDesc);
  
  /**
   * Returns the value of a variable associated with the plugin.
   *
   * (Corresponds to NPP_GetValue.)
   *
   * @param variable - the plugin variable to get
   * @param value - the address of where to store the resulting value
   * @result - NS_OK if this operation was successful
   */
  NS_IMETHOD
  GetValue(nsPluginVariable variable, void *value);
  
  /**
   * Sets the value of a variable associated with the plugin.
   *
   * (Corresponds to NPP_SetValue.)
   *
   * @param variable - the plugin variable to get
   * @param value - the address of the value to store
   * @result - NS_OK if this operation was successful
   */
  NS_IMETHOD
  SetValue(nsPluginVariable variable, void *value);
  
  
  /////////////////////////////////////////////////
  // nsIJVMPlugin
  /////////////////////////////////////////////////

  // This method us used to start the Java virtual machine.
  // It sets up any global state necessary to host Java programs.
  // Note that calling this method is distinctly separate from 
  // initializing the nsIJVMPlugin object (done by the Initialize
  // method).
  NS_IMETHOD
  StartupJVM(nsJVMInitArgs* initargs);
  
  // This method us used to stop the Java virtual machine.
  // It tears down any global state necessary to host Java programs.
  // The fullShutdown flag specifies whether the browser is quitting
  // (PR_TRUE) or simply whether the JVM is being shut down (PR_FALSE).
  NS_IMETHOD
  ShutdownJVM(PRBool fullShutdown = PR_FALSE);
  
  // Causes the JVM to append a new directory to its classpath.
  // If the JVM doesn't support this operation, an error is returned.
  NS_IMETHOD
  AddToClassPath(const char* dirPath);
  
  // Causes the JVM to remove a directory from its classpath.
  // If the JVM doesn't support this operation, an error is returned.
  NS_IMETHOD
  RemoveFromClassPath(const char* dirPath);
  
  // Returns the current classpath in use by the JVM.
  NS_IMETHOD
  GetClassPath(const char* *result);

  NS_IMETHOD
  GetJavaWrapper(JNIEnv* jenv, jint obj, jobject *jobj);
  
  // Find or create a JNIEnv for the current thread.
  // Returns NULL if an error occurs.
  NS_IMETHOD_(nsrefcnt)
    GetJNIEnv(JNIEnv* *result);
  
  // This method must be called when the caller is done using the JNIEnv.
  // This decrements a refcount associated with it may free it.
  NS_IMETHOD_(nsrefcnt)
    ReleaseJNIEnv(JNIEnv* env);

  // This creates a new secure communication channel with Java.
  NS_IMETHOD
  GetSecureJNI(JNIEnv* proxyJNI, nsISecureJNI2* *result);

  // Gives time to the JVM from the main event loop of the browser. This is
  // necessary when there aren't any plugin instances around, but Java threads exist.
  NS_IMETHOD
  SpendTime(PRUint32 timeMillis);
  
protected:
  JavaVM *_vm;
  JNIEnv *_env;
  int _env_refcnt;
  HMonitor _factory_monitor;

  nsISupports *_browser; /* our interface to the browser */
  nsISupports *_console; /* our console interface */
};

#endif /* _japharJVMPlugin_h */
