/*
  This file is part of Japhar, the GNU Virtual Machine for Java Bytecodes.
  Japhar is a project of The Hungry Programmers, GNU, and OryxSoft.

  Copyright (C) 1997, 1998 The Hungry Programmers

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Library General Public
  License as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Library General Public License for more details.

  You should have received a copy of the GNU Library General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  */
#include "config.h"
#include "jni.h"
#include "jniint.h"
#include "native-threads.h"
#include "objects.h"
#include "interp.h"
#include "log.h"
#include "ClazzFile.h"
#include "exceptions.h"

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <assert.h>

#include "op_stack.h"

#define MYLOG "Exceptions"

typedef struct {
  char *classname;
  char *methodname;
  ClazzFile *class;
  MethodStruct *method;
  char *filename;
  u2 pc;
} BackTraceLevel;

/* XXX Hm, are the use of these thread safe? */
static BackTraceLevel backtrace[1000]; /* the stack can't be any deeper than
					  this if you want to throw an
					  exception */

static int backtrace_depth = 0;

/*
 * Map program counter to line number.  Requires the LineNumberBlock
 * to be sorted by increasing pc values.
 */
int
method_pc_to_line_number(JNIEnv *env, MethodStruct *method, u2 pc)
{
  int i;
  assert(NULL != env);
  assert(NULL != method);
  assert(sizeof(u2) < sizeof(int)); /* Need space to return -1 on errors */

  if (method->line_numbers)
    {
      for (i = 0; i < method->num_line_number_blocks; i++)
	{
	  if (pc >= method->line_numbers[i].start_pc)
	    return method->line_numbers[i].line_number;
	}
    }

  /* XXX Unknown line number */
  return -1;
}

void
toplevel_exception_handler(JNIEnv *env,
			   japhar_obj throwable_ref)
{
  int i;
  jclass throwable_class;
  jmethodID getMessageField;
  jstring msg;
  jclass exception_class = (*env)->GetObjectClass(env, throwable_ref);
  ClazzFile *exception_cf = jclass_to_clazzfile(env, exception_class);
  char *exceptionname = getClassName(env, exception_cf);
  HungryJNIEnv *henv = (HungryJNIEnv*)env;
  jthrowable exception_cache;

  throwable_class = (*env)->FindClass(env, "java/lang/Throwable");
  getMessageField = (*env)->GetMethodID(env, throwable_class,
					"getMessage",
					"()Ljava/lang/String;");

  /*
   * Cache exception and make sure the runtime don't think the call to
   * getMessage failed
   */
  exception_cache = henv->_exception;
  henv->_exception = NULL;

  msg = (*env)->CallObjectMethod(env, throwable_ref, getMessageField);

  /* Don't know what to do if the call fails.  Die a horrible death? */
  assert(NULL == henv->_exception);
  henv->_exception = exception_cache;

  if (msg)
    {
      const jbyte *msg_bytes = (*env)->GetStringUTFChars(env, msg, NULL);

      fprintf (stderr, "%s (%s)\n", exceptionname, msg_bytes);

      (*env)->ReleaseStringUTFChars(env, msg, msg_bytes);
    }
  else
    fprintf (stderr, "%s\n", exceptionname);

  for (i = 0; i < backtrace_depth; i ++)
    {
      int line_number = method_pc_to_line_number(env, backtrace[i].method,
					  backtrace[i].pc);;
      
      if (line_number == -1)
	fprintf (stderr, "  in %s.%s%s(pc = %d)\n", backtrace[i].classname,
		 backtrace[i].method->name,
		 SIG_formatToJavaSig(env, backtrace[i].method->sig),
		 backtrace[i].pc);
      else
	fprintf (stderr, "  at %s.%s%s(%d, pc = %d)\n", backtrace[i].classname,
		 backtrace[i].methodname,
		 SIG_formatToJavaSig(env, backtrace[i].method->sig),
		 line_number,
 		 backtrace[i].pc);
    }

  /* now it is safe clear the exception from this thread. */
  (*env)->ExceptionClear(env);

  /* this return will either unwind us to the thread's start function,
     which will return and clean up the thread, or it will return us
     to the toplevel JNI call that started this thread's
     interpreter loop. */
  return;
}

void
cleanup_exception(JNIEnv *env)
{
  int i;

  for (i = 0; i < backtrace_depth; i ++)
    {
       free(backtrace[i].classname);
       free(backtrace[i].methodname);
    }
}

void
throw_exception(JNIEnv *env, japhar_obj throwable_ref, StackFrame *f)
{
  StackFrame *throw_frame, *new_throw_frame;
#ifdef LOG
  jclass throwable_class = (*env)->GetObjectClass(env, throwable_ref);
  ClazzFile *throwable_cf = jclass_to_clazzfile(env, throwable_class);
#endif

  assert(throwable_ref);
  if (!throwable_ref) return;

  JAVARLOG0(MYLOG, 1, "in throw_exception()\n");

  throw_frame = f;

  JAVARLOG4(MYLOG, 1, 
	    "Exception %s thrown from %s.%s - at pc %d\n",
	    getClassName(ENV(throw_frame), throwable_cf),
	    CLASSNAME(throw_frame),
	    METHODNAME(throw_frame),
	    PC(throw_frame));

  while (throw_frame != THREAD_INFO(f)->stack_highwater)
    {
      if (throw_frame->flags & FRAME_NATIVE)
	{
	  JAVARLOG2(MYLOG, 3, "... we hit a native frame (%s.%s)\n",
		    CLASSNAME(throw_frame),
		    METHODNAME(throw_frame));
	  /* if we hit a native frame, we just return.
	     the interpreter loop will return to it's caller
	     if the exception hasn't been handled here. */
	  
	  return;
	}
      else
	{
	  int i;

	  JAVARLOG3(MYLOG, 3, "... we're at a java frame (%s.%s pc %d)\n",
		    CLASSNAME(throw_frame),
		    METHODNAME(throw_frame),
		    PC(throw_frame));

	  for (i = 0; i < NUMEXCEPTIONS(throw_frame); i ++)
	    {
	      ClazzFile *catch_class;
	      
	      if (PC(throw_frame) < EXCEPTIONS(throw_frame)[i].start_pc
		  || PC(throw_frame) > EXCEPTIONS(throw_frame)[i].end_pc)
		continue;

	      catch_class = ExceptionBlock_getHandlerClazz(ENV(throw_frame),
					   METHOD(throw_frame)->clazz,
					   &(EXCEPTIONS(throw_frame)[i]));
	      
	      if (!is_instance_of(env, throwable_ref, catch_class))
		continue;
	      
	      /* we found a match.  reset the pc,
		 push the exception on the stack
		 and return. */
	      PC(throw_frame) = EXCEPTIONS(throw_frame)[i].handler_pc;
	      
	      OPSTACK(throw_frame)->stack_top = OPSTACK_TOP(throw_frame);

	      op_stack_push_object(env, OPSTACK(throw_frame), throwable_ref);
	      
	      (*env)->ExceptionClear(env);

	      backtrace_depth = 0;

	      JAVARLOG4(MYLOG, 1, 
			"Exception %s caught by %s.%s - at pc %d\n",
			getClassName(ENV(throw_frame), catch_class),
			CLASSNAME(throw_frame),
			METHODNAME(throw_frame),
			PC(throw_frame));
	      
	      return;
	    }

	  backtrace[backtrace_depth].classname =
	    strdup(getClassName(env, METHOD(throw_frame)->clazz));
	  backtrace[backtrace_depth].methodname =
	    strdup(METHOD(throw_frame)->name);

	  backtrace[backtrace_depth].method = METHOD(throw_frame);
	  backtrace[backtrace_depth].class = METHOD(throw_frame)->clazz;
	  backtrace[backtrace_depth].pc = PC(throw_frame);

	  backtrace_depth ++;

	  /* if we didn't find a match, pop the stack 
	     frame and do it again. */
	  new_throw_frame = PARENT(throw_frame);
	  pop_frame(THREAD_INFO(throw_frame));
	  throw_frame = new_throw_frame;
	}
    }

  toplevel_exception_handler(env, throwable_ref);
}

void
throw_Exception(JNIEnv *env,
		const char *exception_name,
		const char *msg)
{
  jclass exception = (*env)->FindClass(env, exception_name);
  
  if (!exception)
    {
      (*env)->FatalError(env, "Unable to raise exception.");
    }

  /* NULL == msg means no message */
  (*env)->ThrowNew(env, exception, msg);
}
