/* -*- Mode: C; c-file-style: "gnu" -*-
   init.c -- Startup functions
   Created: Petter Reinholdtsen <pere@td.org.uit.no>, 1998-12-04
 */
/*
  This file is part of Japhar, the GNU Virtual Machine for Java Bytecodes.
  Japhar is a project of The Hungry Programmers, GNU, and OryxSoft.

  Copyright (C) 1997, 1998 The Hungry Programmers

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Library General Public
  License as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Library General Public License for more details.

  You should have received a copy of the GNU Library General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  */

#include "config.h"
#include "init.h"
#include "dynamic_loading.h"
#include "resolve.h"
#include "objects.h"
#include "method.h"
#include "op_stack.h"
#include "interp.h"
#include "exceptions.h"
#include "alloc.h"

/*
 * Return true (1) if everything went OK, and false (0) otherwise.
 */
int
initialize_system_libraries(void)
{
  DLL_LibHandle native_handle = DLL_find("lang");
  if (!native_handle)
    return 0;

  native_handle = DLL_find("io");
  if (!native_handle)
    return 0;

  native_handle = DLL_find("util");
  if (!native_handle)
    return 0;

  native_handle = DLL_find("reflect");
  if (!native_handle)
    return 0;

  return 1;
}

void
initialize_system_threads(JNIEnv *env)
{
  jclass System, ThreadGroup, Thread;
  jobject system_thread_group, system_thread;
  japhar_object *jobj;
  StackFrame *initial_native_frame;
  JThreadInfo *initial_native_thread_info;
  jfieldID field;
  jmethodID method;
  jstring thread_name;
  jint norm_priority;

  initial_native_thread_info = (JThreadInfo*)jcalloc(env, 1,
						     sizeof(JThreadInfo));
  if (NULL == initial_native_thread_info)
    {
      throw_Exception(env, "java/lang/VirtualMachineError",
		      "Cannot allocate thread-local info");
      return;
    }
  initial_native_thread_info->name = jstrdup(env, "main");

  initial_native_thread_info->op_stack = op_stack_allocate(OPSTACK_SIZE);
  if (NULL == initial_native_thread_info->op_stack)
    {
      free(initial_native_thread_info->name);
      free(initial_native_thread_info);
      throw_Exception(env, "java/lang/VirtualMachineError",
		      "Cannot allocate thread-local op_stack");
      return;
    }

  if ( JNI_FALSE == setup_stackframes(initial_native_thread_info) )
    {
      free(initial_native_thread_info->name);
      free(initial_native_thread_info);
      op_stack_deallocate(initial_native_thread_info->op_stack);
      throw_Exception(env, "java/lang/VirtualMachineError",
		      "Cannot allocate thread-local stack");
      return;
    }
  initial_native_thread_info->thread_id = THREAD_getCurrent();

  THREAD_setJavaInfo(initial_native_thread_info);

  System = clazzfile_to_jclass(env, find_class(env, "java/lang/System"));
  Thread = clazzfile_to_jclass(env, find_class(env, "java/lang/Thread"));
  ThreadGroup = clazzfile_to_jclass(env, find_class(env, "java/lang/ThreadGroup"));

  if ((!Thread) || (!ThreadGroup))
    {
      fprintf(stderr,
	      "Unable to initialize threads: cannot find class java/lang/%s\n",
	      ((!Thread) ? "Thread" : "ThreadGroup"));
      exit(1);
    }
  /* just to make sure. */
  if (!System)
    {
      fprintf(stderr,
	      "Unable to initialize vm:  cannot find class java/lang/System\n");
      exit(1);
    }

  system_thread_group = new_object(env, jclass_to_clazzfile(env, ThreadGroup));
  system_thread = new_object(env, jclass_to_clazzfile(env, Thread));

  /* have to put this off until here. */
  initial_native_thread_info->java_thread = system_thread;

  STATE(initial_native_thread_info) = STATE_RUNNING;

  obj_to_object(system_thread_group, jobj);
#if 0
  GC_addReference((JavaVM*)((HungryJNIEnv*)env)->_vm, (GC_obj*)jobj);
#endif

  obj_to_object(system_thread, jobj);
#if 0
  GC_addReference((JavaVM*)((HungryJNIEnv*)env)->_vm, (GC_obj*)jobj);
#endif

#if norm_priority_bogus
  field = (*env)->GetStaticFieldID(env, Thread, "NORM_PRIORITY", "I");
  norm_priority = (*env)->GetStaticIntField(env, Thread, field);
#else
  norm_priority = 10;
#endif
  field = (*env)->GetFieldID(env, Thread, "priority", "I");
  (*env)->SetIntField(env, system_thread, field, norm_priority);

  NSA_SetNativeState(system_thread, initial_native_thread_info);

  initial_native_frame = push_frame(initial_native_thread_info,
				    0);

  OPSTACK_TOP(initial_native_frame) = OPSTACK(initial_native_frame)->stack_top;
  THISPTR(initial_native_frame) = NULL;
#ifdef DEBUG
  METHODNAME(initial_native_frame) = jstrdup(env,"<initial native frame>");
#endif

  initial_native_frame->flags |= FRAME_NATIVE;

  /* we need to initialize the system class first (if it's necessary).
     we also print a warning message since this isn't the classes.zip
     we're expecting to use. */
  method = find_static_method(env, System,
				     "initializeSystemClass",
				     "()V");
  if (method)
    {
      initial_native_frame->method = method;
      (*env)->CallStaticVoidMethod(env, System, method);
    }

  method = find_method(env, ThreadGroup, "<init>", "()V");
  initial_native_frame->method = method;
  (*env)->CallVoidMethod(env, system_thread_group, method);

  thread_name = (*env)->NewStringUTF(env, "main");
  (*env)->NewGlobalRef(env, thread_name);

  method = find_method(env, Thread, "<init>",
		       "(Ljava/lang/ThreadGroup;Ljava/lang/String;)V");
  initial_native_frame->method = method;
  (*env)->CallVoidMethod(env, system_thread, method, system_thread_group, thread_name);

  pop_frame(initial_native_thread_info);
}

/* we have to do this one by hand, since all the class initialization
   stuff relies on java.lang.Class already being initialized. */
void
initialize_class_class(JNIEnv *env)
{
  ClazzFile *class_cf = find_class(env, "java/lang/Class");

  class_cf->num_fields = 0;
  class_cf->nesting_level = 1;

  calculate_instance_field_offsets(env, class_cf);
}
