/* -*- Mode: C; c-file-style: "gnu" -*-
   nativeglue.c -- glue for calling between Java<=>C.
   Created: Chris Toshok <toshok@hungry.com>
 */
/*
  This file is part of Japhar, the GNU Virtual Machine for Java Bytecodes.
  Japhar is a project of The Hungry Programmers, GNU, and OryxSoft.

  Copyright (C) 1997, 1998 The Hungry Programmers

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Library General Public
  License as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Library General Public License for more details.

  You should have received a copy of the GNU Library General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  */
#include "config.h"
#include "qop.h"
#include "jniint.h"
#include "log.h"
#include "dynamic_loading.h"
#include "exceptions.h"
#include "interp.h"
#include "sig.h"
#include "ClazzFile.h"
#include "nativeglue.h"
#include "compat.h"
#include "primitive-class.h"
#include "util.h"

#ifdef HAVE_LIBFFI
#include "ffi.h"
#endif

#include <stdio.h>
#include <string.h>
#include <signal.h>
#include <stdlib.h>
#include <assert.h>

#include "op_stack.h"

#define MYLOG "NativeGlue"

/*
 * Convert (any/class/Name, method) to Java_any_class_Name_method.
 */
static void
create_short_native_name(char *classname,
			 MethodStruct *method,
			 char *func_name,
			 int func_name_length)
{
  int ret_val;

  ret_val = snprintf(func_name, func_name_length,
#ifdef HAVE_LIBFFI
		     "Java_%s_%s",
#else
		     "Java_%s_%s_stub",
#endif
		     classname, method->name);

  if (-1 == ret_val)
    { /* XXX Output was truncated.  What should we do? */
      fprintf(stderr, "Error: Class method %s.%s truncated.\n",
	      classname, method->name);
    }

  slashes_to_underscores(func_name);
}


/*
 * Convert (any/class/Name, method) to Java_any_class_Name_method__sig.
 */
static void
create_long_native_name(JNIEnv *env, char *classname,
			MethodStruct *method,
			char *func_name,
			int func_name_length)
{
  int retval;
  char *hacked_sig = SIG_formatToNativeName(env, method->sig);

  retval = snprintf(func_name, func_name_length,
#ifdef HAVE_LIBFFI
		    "Java_%s_%s__%s",
#else
		    "Java_%s_%s__%s_stub",
#endif
		    classname, method->name, hacked_sig);
  free(hacked_sig);

  if (-1 == retval)
    { /* XXX Output was truncated.  What should we do? */
      fprintf(stderr, "Error: Class method %s.%s truncated.\n",
	      classname, method->name);
    }

  slashes_to_underscores(func_name);
}

static void
resolve_native_func(JNIEnv *env, MethodStruct *method)
{
  char native_func_name[1024];
  char *class_name = getClassName(env, method->clazz);
  
  create_short_native_name(class_name, method,
			   native_func_name, sizeof(native_func_name));

  JAVARLOG1(MYLOG, 1, "   short native stub name is %s()\n", native_func_name);
  
  method->native_func = (NativeFunc)DLL_findFunction(native_func_name);

  if (method->native_func == NULL)
    {
      /* now we try the long name. */
      create_long_native_name(env, class_name, method,
			      native_func_name, sizeof(native_func_name));

      method->native_func = (NativeFunc)DLL_findFunction(native_func_name);
    }

  if (method->native_func == NULL) /* what do we do here? */
    {
      char *msg;
      
      msg = (char*)malloc(sizeof(char) 
			  * (strlen("Unresolved Function: ")
			     + strlen(native_func_name) + 1));
      assert(NULL != msg);
      sprintf(msg, "Unresolved Function: %s", native_func_name);
      
      (*env)->FatalError(env, msg);
    }
}

#ifdef HAVE_LIBFFI
static ffi_type *
sig_to_ffi_type(Signature *sig)
{
  switch(sig->any.tag)
    {
    case SIG_PRIM:
      {
	switch (sig->prim.type)
	  {
	  case SIG_JBOOLEAN:
	    return &ffi_type_uint8;
	  case SIG_JBYTE:
	    return &ffi_type_sint8;
	  case SIG_JCHAR:
	    return &ffi_type_uint16;
	  case SIG_JSHORT:
	    return &ffi_type_sint16;
	  case SIG_JINT:
	    return &ffi_type_sint32;
	  case SIG_JFLOAT:
	    return &ffi_type_float;
	  case SIG_JDOUBLE:
	    return &ffi_type_double;
	  case SIG_JLONG:
	    return &ffi_type_sint64;
	  case SIG_JVOID:
	    return &ffi_type_void;
	  case SIG_JOBJECT:
	    return &ffi_type_pointer;
	  default:
	    assert(0);
	    return NULL;
	  }
      }
    case SIG_CLASS:
      /* fallthrough */
    case SIG_ARRAY:
      return &ffi_type_pointer;
    default:
      assert(0);
      return NULL;
    }
}

static void*
sig_to_jvalue_addr(Signature *sig,
		   jvalue *v)
{
  switch (sig->any.tag)
    {
    case SIG_PRIM:
      {
	switch (sig->prim.type)
	  {
	  case SIG_JBOOLEAN:
	    return &v->z;
	  case SIG_JBYTE:
	    return &v->b;
	  case SIG_JCHAR:
	    return &v->c;
	  case SIG_JSHORT:
	    return &v->s;
	  case SIG_JINT:
	    return &v->i;
	  case SIG_JFLOAT:
	    return &v->f;
	  case SIG_JDOUBLE:
	    return &v->d;
	  case SIG_JLONG:
	    return &v->j;
	  case SIG_JVOID:
	    return NULL; /* XXX */
	  case SIG_JOBJECT:
	    return &v->l;
	  default:
	    assert(0);
	    return NULL;
	  }
      }
    case SIG_CLASS:
    case SIG_ARRAY:
      return &v->l;
    default:
      assert(0);
      return NULL;
    }
}
#endif

static jvalue
do_native_method_call_with_args(JNIEnv *env,
				MethodStruct *method,
				jvalue *args)
{
  HungryJNIEnv *henv = (HungryJNIEnv*)env;
  StackFrame *new_frame = push_frame(henv->_java_info, 0);

  new_frame->flags |= FRAME_NATIVE;
  METHOD(new_frame) = method;
#ifdef DEBUG
  CLASSNAME(new_frame) = getClassName(env, method->clazz);
  METHODNAME(new_frame) = strdup (METHOD(new_frame)->name);
  assert(NULL != METHODNAME(new_frame));
#endif

  /* get the function pointer for this native method, if we don't already
     know it. */
  if (!method->native_func)
    resolve_native_func(env, method);

  OPSTACK_TOP(new_frame) = OPSTACK(new_frame)->stack_top;

  if (method->access_flags & ACC_STATIC)
    THISPTR(new_frame) = NULL;
  else
    THISPTR(new_frame) = args[1].l;

  maybe_enter_monitor_for_method(ENV(new_frame),
				 METHOD(new_frame),
				 THISPTR(new_frame));

#ifdef HAVE_LIBFFI
  {
    MethodStruct *method = METHOD(new_frame);
    ffi_type* rtype = sig_to_ffi_type(method->sig->method.return_type);
    ffi_type** atypes;
    ffi_cif cif;
    void *rvalue;
    void **params;
    int i;
    int num_params = SIG_numParams(env, method->sig);

    params = (void**)malloc(sizeof(void*) * (num_params + 2));
    atypes = (ffi_type**)malloc(sizeof(ffi_type*) * (num_params + 2));

    for (i = 2; i < 2 + num_params; i++)
      {
	atypes[ i ] = sig_to_ffi_type(method->sig->method.params[ i - 2 ]);
	params[ i ] = sig_to_jvalue_addr(method->sig->method.params[ i - 2 ],
					 &args[ i ]);
      }

    /* 
    ** these never change.  
    **
    ** parameter 0 is the JNIEnv*
    **
    ** parameter 1 is either the object (for nonstatic
    ** methods) or the classes. 
    */
    atypes[ 0 ] = &ffi_type_pointer;
    params[ 0 ] = (void*)&env;
    atypes[ 1 ] = &ffi_type_pointer;
    params[ 1 ] = &args[ 1 ].l;

    if (ffi_prep_cif(&cif, FFI_DEFAULT_ABI,
		     num_params + 2,
		     rtype, atypes) != FFI_OK)
      {
	assert(0);
      }

    rvalue = sig_to_jvalue_addr(method->sig->method.return_type, &args[0]);

    ffi_call(&cif,
	     FFI_FN(method->native_func),
	     rvalue, params);

    free(params);
    free(atypes);
  }
#else
  (*METHOD(new_frame)->native_func)(env, args);
#endif
  
  if (!SIG_isVoid(env, METHOD(new_frame)->sig->method.return_type))
    {
      switch(METHOD(new_frame)->sig->method.return_type->any.tag)
	{
	case SIG_PRIM:
	  THREAD_INFO(new_frame)->return_value.tag =
	    METHOD(new_frame)->sig->method.return_type->prim.type;
	  break;
	case SIG_ARRAY:
	case SIG_CLASS:
	  THREAD_INFO(new_frame)->return_value.tag = SIG_JOBJECT;
	  break;
	case SIG_METHOD:
	  assert(0); /* this shouldn't ever happen. */
	  break;
	}
    }
  else
    {
      THREAD_INFO(new_frame)->return_value.tag = SIG_JVOID;
    }

  THREAD_INFO(new_frame)->return_value.value = args[0];

  maybe_push_return_value(new_frame);
  
  /* then pop the frame off */
  pop_frame(THREAD_INFO(new_frame));

  return THREAD_INFO(new_frame)->return_value.value;
}

void
do_native_method_call(JNIEnv *env, MethodStruct *method)
{
  HungryJNIEnv *henv = (HungryJNIEnv*)env;
  int num_arguments = 0;
  jvalue args[100];
  int i;
#ifdef LOG
  char *class_name = getClassName(env, method->clazz);
#endif
  StackFrame *current_frame = TOPFRAME(henv->_java_info);

  JAVARLOG2(MYLOG, 1, "Inside do_native_method_call(%s.%s)\n",
	    class_name, method->name);

  initialize_class(env, method->clazz);

  num_arguments = SIG_numParams(env, method->sig);

  for (i = num_arguments + 1;
       i > 1;
       i --)
    {
      op_stack_pop_value(env, OPSTACK(current_frame),
			 (method->sig->method.params[ i - 2  ]->any.tag == SIG_PRIM
			  ? method->sig->method.params[ i - 2 ]->prim.type
			  : SIG_JOBJECT),
			 &args[ i ]);
    }
  
  if (method->access_flags & ACC_STATIC)
    {
      /* fill in the class pointer */
      args[1].l = clazzfile_to_jclass(env, method->clazz);
    }
  else
    {
      /* pop the this pointer */
      op_stack_pop_value(env, OPSTACK(current_frame),
			 SIG_JOBJECT,
			 &args[ 1 ]);
    }

  do_native_method_call_with_args(env, method, args);
}

/* This is the generic method used by the two specific below. */
static jvalue
call_java_method(JNIEnv *env,
		 MethodStruct *method,
		 jobject obj, /* NULL for static calls */
		 jvalue *args)
{
  StackFrame *new_frame;
  HungryJNIEnv *henv = (HungryJNIEnv*)env;

  JAVARLOG2(MYLOG, 4, "calling %s.%s()\n",
	    method->clazz->class_name,
	    method->name);

  initialize_class(env, method->clazz);

  new_frame = create_frame_for_method(henv->_java_info, method);
  if (NULL == new_frame)
    {
      jvalue value;
      /*
       * Fill it with zero.  Assume jlong is the longest type.  I'm
       * not sure what this method should return on error, but...
       */
      value.j = 0;
      return value; /* create_frame_for_methods has already throwed */
    }

  fill_local_vars(new_frame, method, args, obj);

  OPSTACK_TOP(new_frame) = OPSTACK(new_frame)->stack_top;

  THREAD_INFO(new_frame)->return_value.value.j = 0;

  maybe_enter_monitor_for_method(env, method, obj);

  interp_loop(new_frame);

  OPSTACK(new_frame)->stack_top = OPSTACK_TOP(new_frame);

  return THREAD_INFO(new_frame)->return_value.value;
}
  
jvalue
CallJavaMethod (JNIEnv *env,
		MethodStruct *method,
		jobject obj,
		jvalue *args)
{
  japhar_obj real_obj;

  /* in the case of interfaces, we have to do a second lookup when
     we actually call the method, to find the real method we should
     be calling. */
  if (method->clazz->access_flags & ACC_INTERFACE)
    {
      ClazzFile *cf;

      method = get_interface_method_info(env, obj,
					 &cf,
					 method->name, method->sig);

      real_obj = cast_obj((japhar_obj)obj,
			  cf);
    }
  else
    {
      real_obj = cast_obj((japhar_obj)obj,
			  method->clazz);
    }

  assert(real_obj);
  if(!real_obj)
    {
      /* this should be an exception, but I don't know what... */
      jvalue foo;

      foo.l = 0;
      return foo;
    }

  if (method->access_flags & ACC_NATIVE)
    {
      jvalue new_args[100];
      int num_arguments = SIG_numParams(env, method->sig);

      if (num_arguments > 1)
	memcpy(&new_args[2], &args[1], num_arguments);

      new_args[1].l = real_obj;

      return do_native_method_call_with_args(env,
					     method, new_args);
    }
  else
    {
      return call_java_method(env, method, real_obj, args);
    }
}

jvalue
CallStaticJavaMethod (JNIEnv *env,
		      MethodStruct *method,
		      jvalue *args)
{
  if (method->access_flags & ACC_NATIVE)
    {
      jvalue new_args[100];
      int num_arguments = SIG_numParams(env, method->sig);

      if (num_arguments > 0)
	memcpy(&new_args[2], &args[1], num_arguments);

      new_args[1].l = 0;

      return do_native_method_call_with_args(env,
					     method, new_args);
    }
  else
    {
      return call_java_method(env, method, 0, args);
    }
}

/****************************************************************************
 * These are used by libjni/src/jninativ.c to register native method
 * function pointers.
 */
void
register_native_func(JNIEnv *env, MethodStruct *method, void *fnPtr)
{
  /*
   * XXX Should probably save the old function to be able to restore
   * previous state
   */
  method->native_func = fnPtr;
}
void
unregister_native_func(JNIEnv *env, MethodStruct *method)
{
  /*
   * XXX Should probably restore to old value
   */
  method->native_func = NULL;
}

jclass
sig_to_jclass(JNIEnv *env, Signature *sig)
{
  jclass sig_cls = NULL;

  if (sig->any.tag == SIG_CLASS
      || sig->any.tag == SIG_ARRAY)
    {
      char *java_sig;
	  java_sig = SIG_formatToJavaSig(env, sig);
	  
      sig_cls = (*env)->FindClass(env, java_sig);
	  
      free(java_sig);
    }
  else /* SIG_PRIM */
    {
      switch (sig->prim.type)
	{
	case SIG_JBOOLEAN:
	  sig_cls = createFakePrimitiveClass(env, "boolean");
	  break;
	case SIG_JBYTE:
	  sig_cls = createFakePrimitiveClass(env, "byte");
	  break;
	case SIG_JCHAR:
	  sig_cls = createFakePrimitiveClass(env, "char");
	  break;
	case SIG_JSHORT:
	  sig_cls = createFakePrimitiveClass(env, "short");
	  break;
	case SIG_JINT:
	  sig_cls = createFakePrimitiveClass(env, "int");
	  break;
	case SIG_JLONG:
	  sig_cls = createFakePrimitiveClass(env, "long");
	  break;
	case SIG_JFLOAT:
	  sig_cls = createFakePrimitiveClass(env, "float");
	  break;
	case SIG_JDOUBLE:
	  sig_cls = createFakePrimitiveClass(env, "double");
	  break;
	case SIG_JVOID:
	  sig_cls = createFakePrimitiveClass(env, "void");
	  break;
	default:
	  (*env)->FatalError(env, "Illegal primitive type.");
	  break;
	}
    }

  return sig_cls;
}

/**
 * Convert from a methodID into a Java object of class
 * java.lang.reflect.Method or java.lang.reflect.Constructor
 * (whichever corresponds).  The only difference with the JDK 1.2 JNI
 * ToReflectedMethod function is that by using this method,
 * repetitive string lookups can be avoided when inside loops.
 */
jobject
to_reflected_method(JNIEnv *env, jmethodID methodID,
    jclass class_class, jclass method_class, jmethodID method_ctor,
    jfieldID clazz_field, jfieldID slot_field, jfieldID name_field,
    jfieldID returnType_field, jfieldID parameterTypes_field,
    jfieldID exceptionTypes_field)
{
  MethodStruct *method = (MethodStruct *) methodID;
  ClazzFile *clazz = method->clazz;
  jobject rmethod = (*env)->NewObject(env, method_class, method_ctor);
  int num_params = method->sig->method.num_params;
  int num_exceptions = method->num_throwable_exceptions;
  jobjectArray exception_array;
  jobjectArray param_array;
  int p;

  /*
   * We use the NULL-ness of name_field to determine if we are
   * working with a Constructor or a Method
   */
  if (name_field != NULL) {
    jstring mname = (*env)->NewStringUTF(env, method->name);
    jclass returntype = sig_to_jclass(env,
        method->sig->method.return_type);
    (*env)->SetObjectField(env, rmethod, returnType_field,
			   returntype);
    mname = (*env)->NewGlobalRef(env, mname);
    (*env)->SetObjectField(env, rmethod, name_field, mname);
  }

  /* Create the declared exceptions array */
  exception_array = (*env)->NewObjectArray(env, num_exceptions,
      class_class, NULL);
  for (p = 0; p < num_exceptions; p ++)
  {
    jobject exc;
    ClazzFile *exc_type;
  
    exc_type = getThrowableException(env, method, p);
  
    exc = clazzfile_to_jclass(env, exc_type);
  
    (*env)->SetObjectArrayElement(env, exception_array, p, exc);
  }
  exception_array = (*env)->NewGlobalRef(env, exception_array);
  
  /* Create the parameter types array */
  param_array = (*env)->NewObjectArray(env, num_params,
      class_class, NULL);
  for (p = 0; p < num_params; p ++)
  {
    jclass param;
  
    param = sig_to_jclass(env, method->sig->method.params[p]);

    (*env)->SetObjectArrayElement(env, param_array, p, param);
  }
  param_array = (*env)->NewGlobalRef(env, param_array);
  
  NSA_SetNativeState(rmethod, methodID);
  (*env)->SetObjectField(env, rmethod, clazz_field,
			 clazzfile_to_jclass(env, clazz));
  (*env)->SetObjectField(env, rmethod, exceptionTypes_field,
			 exception_array);
  (*env)->SetObjectField(env, rmethod, parameterTypes_field,
			 param_array);
  return (*env)->NewGlobalRef(env, rmethod);
}

/**
 * Convert from a fieldID into a Java object of class
 * java.lang.reflect.Field.  The only difference with the JDK 1.2 JNI
 * ToReflectedField function is that by using this method,
 * repetitive string lookups can be avoided when inside loops.
 */
jobject
to_reflected_field(JNIEnv *env, jfieldID fieldID,
    jclass field_class, jmethodID field_ctor,
    jfieldID slot_field, jfieldID type_field, jfieldID name_field,
    jfieldID clazz_field)
{
  jobject rfield;
  jstring name;
  jclass type;
  FieldStruct *field = (FieldStruct *) fieldID;
  ClazzFile *clazz = field->clazz;

  rfield = (*env)->NewObject(env, field_class, field_ctor);
  name = (*env)->NewStringUTF(env, field->name);
  type = sig_to_jclass(env, field->sig);

  name = (*env)->NewGlobalRef(env, name);
  (*env)->SetObjectField(env, rfield, name_field, name);
  NSA_SetNativeState(rfield, fieldID);
  (*env)->SetObjectField(env, rfield, clazz_field,
			 clazzfile_to_jclass(env, clazz));
  (*env)->SetObjectField(env, rfield, type_field, type);
  return (*env)->NewGlobalRef(env, rfield);
}
